from typing import List

import pandas as pd
import torch.nn as nn

from .base_nn import BaseNetwork
from ..retrieve_keys import get_model_keys
from ..args_val import (is_positive_int, is_iterable, has_activation_functions,
                        activation_functions_check)


class FNN(BaseNetwork):
    """
    A Feedforward Neural Network (FNN) model for supervised learning.

    References:
        - Suganthan, P. N., & Katuwal, R. (2021). On the origins of randomization-based feedforward neural networks.
          *Applied Soft Computing*, 105, 107239. [DOI: 10.1016/j.asoc.2021.107239](https://doi.org/10.1016/j.asoc.2021.107239)

    """

    def __init__(self,
                 tabular: dict | pd.DataFrame,
                 visualise: bool = False) -> None:
        """
        Class description goes here
        :param tabular: An input accepting both a dictionary or a pandas.DataFrame object.
        :param visualise: A toggle switch to visualize the model. OFF(False) by default.
        """

        super().__init__(tabular, get_model_keys("FNN"), visualise)

    def _build_model(self, input_size: int, output_size: int,
                     hidden_sizes: List[int],
                     activation_functions: List[nn.Module]) -> nn.Sequential:
        """
        A procedural function declaring the steps of building a model.
        :param input_size: The number of input features.
        :param output_size: The number of output features.
        :param hidden_sizes: The number of nodes in each hidden layer.
        :param activation_functions: The activation functions to use.
        :return: A torch.nn.Sequential object representing the layers.
        """

        is_positive_int(input_size)
        is_positive_int(output_size)
        is_iterable(hidden_sizes)

        for sizes in hidden_sizes:
            is_positive_int(sizes)

        has_activation_functions(activation_functions)
        activation_functions_check(activation_functions, hidden_sizes)

        return nn.Sequential(*self._create_layers(
            input_size, output_size, hidden_sizes, activation_functions))

    def _create_layers(
            self, input_size: int, output_size: int, hidden_sizes: List[int],
            activation_functions: List[nn.Module]) -> list[nn.Module]:
        """
        A function to generate layers dynamically.
        :param input_size: The number of input features.
        :param hidden_sizes: The number of nodes in each hidden layer.
        :param output_size: The number of output features.
        :param activation_functions: The activation functions to use.
        :return: A list of torch.nn activation functions. Refer to the official documentation for possible inputs:
        https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity and
        https://pytorch.org/docs/stable/nn.html#non-linear-activations-other
        """

        layers = []

        in_size = input_size

        for hidden_size, activation_function in zip(hidden_sizes,
                                                    activation_functions):
            layers += [nn.Linear(in_size, hidden_size), activation_function]

            # go to next layer
            in_size = hidden_size

        layers.append(nn.Linear(hidden_sizes[-1], output_size))
        return layers



import torch.nn as nn
from template_nn import CNN


def create_cnn_model():
    # Example of creating a Convolutional Neural Network (CNN)
    # This CNN is configured for image input (e.g., 3 channels, 64x64 pixels).
    # It includes two convolutional layers and two fully connected layers.
    cnn_model = CNN({
        "image_size": (64, 64),
        "conv_channels":
        [3, 6, 16],  # Input channels, then output channels for each conv layer
        "conv_kernel_size": 3,
        "pool_kernel_size": 2,
        "fcn_hidden_sizes": [120,
                             80],  # Hidden layers for the fully connected part
        "activation_functions":
        [nn.ReLU(), nn.ReLU()],  # Activations for conv and fc layers
        "output_channel": 10  # Output classes
    })
    print("CNN Model created successfully:")
    print(cnn_model)
    # Example of a training loop (simplified)
    # optimizer = torch.optim.Adam(cnn_model.parameters(), lr=0.001)
    # criterion = nn.CrossEntropyLoss()
    # for epoch in range(num_epochs):
    #     outputs = cnn_model(inputs)
    #     loss = criterion(outputs, labels)
    #     optimizer.zero_grad()
    #     loss.backward()
    #     optimizer.step()
    return cnn_model


if __name__ == "__main__":
    create_cnn_model()



import json
import os

# Path to the JSON file containing model keys
KEY_FILE_PATH = os.path.join(os.path.dirname(__file__), "keys.json")


def get_model_keys(model_name: str) -> list:
    """
    Retrieve the configuration keys for the given model name from keys.json.

    :param model_name: Name of the model
    :return: List of expected configuration keys
    :raises FileNotFoundError: If the keys.json file does not exist
    :raises ValueError: If the model name is not found in keys.json
    """
    if not os.path.exists(KEY_FILE_PATH):
        raise FileNotFoundError(f"keys.json not found at {KEY_FILE_PATH}")

    with open(KEY_FILE_PATH, "r") as f:
        keys_data = json.load(f)

    if model_name not in keys_data:
        raise ValueError(f"Model name '{model_name}' not found in keys.json")

    return keys_data[model_name]



from typing import List
import pandas as pd
import torch.nn as nn

from ..args_val import is_iterable, is_positive_int

from .base_nn import BaseNetwork
from .cml import CML
from .fnn import FNN
from ..retrieve_keys import get_model_keys


class CNN(BaseNetwork):
    """
    A Convolutional Neural Network (CNN) model for supervised learning.
    """

    def __init__(self,
                 tabular: dict | pd.DataFrame,
                 visualise: bool = False) -> None:
        """
        Class description goes here
        :param tabular: An input accepting both a dictionary or a pandas.DataFrame object.
        :param visualise: A toggle switch to visualize the model. OFF(False) by default.
        """

        super().__init__(tabular, get_model_keys("CNN"), visualise)
        self.image_size = (0, 0)

    def _build_model(self, image_size: tuple[int, int],
                     conv_channels: List[int], conv_kernel_size: int,
                     pool_kernel_size: int, fcn_hidden_sizes: List[int],
                     activation_functions: List[nn.Module],
                     output_channel: int) -> nn.Sequential:

        self.image_size = image_size
        is_iterable(self.image_size)

        for sizes in self.image_size:
            is_positive_int(sizes)

        return nn.Sequential(*self._create_layers(
            conv_channels, conv_kernel_size, pool_kernel_size,
            fcn_hidden_sizes, activation_functions, output_channel))

    def _create_layers(self, conv_channels: List[int], conv_kernel_size: int,
                       pool_kernel_size: int, fcn_hidden_sizes: List[int],
                       activation_functions: List[nn.Module],
                       output_channel: int) -> List[nn.Module]:

        height, width = self.image_size

        for _ in range(len(conv_channels) - 1):
            height, width = self._compute_output_dim(height, width,
                                                     conv_kernel_size,
                                                     pool_kernel_size)

        conv_layers = CML(
            {
                "conv_channels": conv_channels,
                "conv_kernel_size": conv_kernel_size,
                "pool_kernel_size": pool_kernel_size
            },
            visualise=False)

        fcn_layers = FNN(
            {
                "input_size": conv_channels[-1] * height * width,  # temp
                "hidden_sizes": fcn_hidden_sizes,
                "output_size": output_channel,
                "activation_functions": activation_functions
            },
            visualise=False)

        return [conv_layers, nn.Flatten(), fcn_layers]

    def _compute_dim(self,
                     in_dim: int,
                     kernel_size: int,
                     stride=1,
                     padding=0,
                     dilation=1) -> int:
        return (in_dim + 2 * padding - dilation *
                (kernel_size - 1) - 1) // stride + 1

    def _compute_output_dim(self, height: int, width: int,
                            conv_kernel_size: int,
                            pool_kernel_size: int) -> tuple[int, int]:

        height = self._compute_dim(height, conv_kernel_size)
        width = self._compute_dim(width, conv_kernel_size)

        height = self._compute_dim(height,
                                   pool_kernel_size,
                                   stride=pool_kernel_size)
        width = self._compute_dim(width,
                                  pool_kernel_size,
                                  stride=pool_kernel_size)

        return (height, width)



from typing import List
import torch.nn as nn

from ..args_val import (is_iterable, is_positive_int)

from ..retrieve_keys import get_model_keys
from .base_nn import BaseNetwork


class CML(BaseNetwork):

    def __init__(self, tabular, visualise):
        super().__init__(tabular, get_model_keys("CML"), visualise)

    def _build_model(self,
                     conv_channels: List[int],
                     conv_kernel_size: int = 3,
                     pool_kernel_size: int = 3) -> nn.Sequential:

        is_positive_int(conv_kernel_size)
        is_positive_int(pool_kernel_size)

        is_iterable(conv_channels)
        for channel_sizes in conv_channels:
            is_positive_int(channel_sizes)

        return nn.Sequential(*self._create_layers(
            conv_channels, conv_kernel_size, pool_kernel_size))

    def _create_layers(self,
                       conv_channels: List[int],
                       conv_kernel_size: int = 3,
                       pool_kernel_size: int = 3) -> List[nn.Module]:

        layers = []

        in_size = conv_channels[0]

        for out_size in conv_channels[1:]:
            layers += [
                nn.Conv2d(
                    in_channels=in_size,
                    out_channels=out_size,
                    kernel_size=conv_kernel_size,
                ),
                nn.MaxPool2d(kernel_size=pool_kernel_size, stride=2)
            ]

            in_size = out_size

        return layers



import torch.nn as nn
from template_nn import FNN


def create_fnn_model():
    # Example of creating a Feedforward Neural Network (FNN)
    # This FNN has an input size of 10, output size of 2, and two hidden layers of sizes 64 and 32.
    # ReLU activation functions are used for both hidden layers.
    fnn_model = FNN({
        "input_size": 10,
        "output_size": 2,
        "hidden_sizes": [64, 32],
        "activation_functions": [nn.ReLU(), nn.ReLU()]
    })
    print("FNN Model created successfully:")
    print(fnn_model)
    # Example of a training loop (simplified)
    # optimizer = torch.optim.Adam(fnn_model.parameters(), lr=0.001)
    # criterion = nn.CrossEntropyLoss()
    # for epoch in range(num_epochs):
    #     outputs = fnn_model(inputs)
    #     loss = criterion(outputs, labels)
    #     optimizer.zero_grad()
    #     loss.backward()
    #     optimizer.step()
    return fnn_model


if __name__ == "__main__":
    create_fnn_model()



import torch
import torch.nn as nn

from ..args_val import is_df, is_dict, is_valid_keys


class BaseNetwork(nn.Module):

    def __init__(self, tabular, model_keys, visualise) -> None:
        super().__init__()
        self.tabular = tabular
        self.model_keys = model_keys
        self.visualise = visualise
        self.model = self._build_model(*self._get_params(tabular, model_keys))

        print(self) if visualise else None

    def forward(self, x: torch.Tensor) -> nn.Module:
        return self.model(x)

    def _build_model(self, *args, **kwargs):
        raise NotImplementedError("Define how model is built here")

    def _create_layers(self, *args, **kwargs):
        raise NotImplementedError("Define layer structure here")

    def _get_params(self, tabular, model_keys) -> list:
        """
        Destructures a tabular input.
        :param keys: A tuple containing keys for specific use case.
        :param tabular: A dict or pd.DataFrame input.
        :return: A tuple containing values relevant to the `keys` list.
        """
        is_valid_keys(tabular, model_keys)

        return (is_dict(tabular, model_keys)
                if isinstance(tabular, dict) else is_df(tabular, model_keys))

    def optimise(self):
        pass



from typing import Iterable, List
import numpy as np
import pandas as pd
from torch.nn import Module


def is_valid_keys(tabular: dict | pd.DataFrame, keys: tuple) -> None:
    if not all(key in tabular for key in keys):
        raise ValueError(f"Tabular data must contain keys {keys}")


def is_dict(tabular: dict | pd.DataFrame, keys: tuple) -> list:
    params = []
    for key in keys:
        params.append(tabular[key])
    return params


def is_df(tabular: dict | pd.DataFrame, keys: tuple) -> list:
    params = []
    for key in keys:
        value = tabular[key].iloc[0]

        if isinstance(value, np.integer):
            params.append(value.item())
        else:
            params.append(value)
    return params


def is_positive_int(number: int) -> None:
    if not isinstance(number, int) or number <= 0:
        raise ValueError(f"Expected positive integer, got {number} instead.")


def is_iterable(numbers: Iterable[int] | List[tuple[int, int]]) -> None:
    if not isinstance(numbers, Iterable):
        raise ValueError(
            f"Expected iterable structure, got {numbers} instead.")


def has_activation_functions(activation_functions: Iterable[Module]) -> None:
    if not activation_functions:
        raise ValueError("No activation functions were provided.")


def activation_functions_check(activation_functions: List[Module],
                               hidden_sizes: List[int]) -> None:
    if len(activation_functions) != len(hidden_sizes):
        raise ValueError(
            "Number of activation functions does not match number of hidden node counts."
            +
            f"Expected {activation_functions} of hidden nodes, but got {hidden_sizes} instead."
        )
